// Toolkit.cpp
// Neuron Computing Inc.

#include "Toolkit.h"

#include <stdio.h>
#include <X11/IntrinsicP.h> 

#ifndef MOTIF
const int y_offset = 18; // Make room for the menu button
#else
const int y_offset = 28; // Make room for the menu button
#endif

#ifdef MOTIF
	#include <Xm\MainW.h>
#else
#endif
// Global
extern XApp* g_app;

// Global Component Event Handlers
void component_mouseDown(Widget w, Component* comp, XEvent* event, Boolean* cont)
{
	comp->mouseDown(event->xbutton.x, event->xbutton.y - y_offset);
}

void component_mouseMove(Widget w, Component* comp, XEvent* event, Boolean* cont)
{
	comp->mouseMove(event->xbutton.x, event->xbutton.y - y_offset);
}

void component_mouseUp(Widget w, Component* comp, XEvent* event, Boolean* cont)
{
	comp->mouseUp(event->xbutton.x, event->xbutton.y - y_offset);
}

void component_paint(Widget w , Component* comp, XEvent* event, Boolean* cont)
{
	Graphics* g = comp->getGraphics();
	comp->update(g);
	*cont = FALSE;
}

//////////////////////////////////////////////////////////////////
// Graphics
Graphics::Graphics()
{
	m_gc = NULL;
}

Graphics::~Graphics()
{
	if (m_gc != NULL)
		XtReleaseGC(m_w, m_gc);
}

void Graphics::setColor(Color c)
{
	XSetForeground(XtDisplay(m_w), m_gc, c.m_rgb);
}

void Graphics::drawLine(int x1, int y1, int x2, int y2)
{
	// Draw once to clear the previous line.
	XDrawLine(XtDisplay(m_w), XtWindow(m_w), m_gc, 
				x1, y1, x2, y2);
}

void Graphics::drawRect(int x, int y, int width, int height)
{
	XDrawRectangle(XtDisplay(m_w), XtWindow(m_w), m_gc, 
					x, y, width, height);
}

void Graphics::fillRect(int x, int y, int width, int height)
{
	XGCValues v;
	unsigned long valuemask = GCForeground | GCBackground | GCFont;
	XGetGCValues(XtDisplay(m_w), m_gc, valuemask, &v);
	setColor(v.background);
	XFillRectangle(XtDisplay(m_w), XtWindow(m_w), m_gc, 
					x, y, width-1, height-1);
	setColor(v.foreground);
}

void Graphics::drawString(int x, int y, char *str)
{
	// Draw once to clear the previous line.
	XDrawString(XtDisplay(m_w), XtWindow(m_w), m_gc,
				x, y, str, strlen(str));
}

void Graphics::scroll(int x, int y, int width, int height, int dx, int dy)
{
	XCopyArea(XtDisplay(m_w), XtWindow(m_w), XtWindow(m_w), m_gc,
				// Source
				x, y, width, height,
				// Destination
				dx, dy);
}

FontMetrics* Graphics::getFontMetrics()
{
	return new FontMetrics;
}

//////////////////////////////////////////////////////////////////
// Component
Component::Component()
{
	m_w = NULL;
	m_g = NULL;
}

Component::~Component()
{
	if (m_g != NULL)
		delete m_g;
	if (m_w != NULL)
		XtDestroyWidget(m_w);
}

void Component::registerEventHandlers(Widget w)
{
	XtAddEventHandler(w, ExposureMask, 1, (XtEventHandler)component_paint,
						(XtPointer)this);
	XtAddEventHandler(w, ButtonPressMask, FALSE,
						(XtEventHandler)component_mouseDown, (XtPointer)this);
	XtAddEventHandler(w, ButtonMotionMask, FALSE,
						(XtEventHandler)component_mouseMove, (XtPointer)this);
	XtAddEventHandler(w, ButtonReleaseMask,
						FALSE, (XtEventHandler)component_mouseUp, (XtPointer)this);
}

void Component::setBounds(int x, int y, int width, int height)
{
	Widget w = m_w;
	int borderWidth = 0;
	XtConfigureWidget(w, x, y, width, height, borderWidth);
}

Rectangle* Component::getBounds()
{
	Widget w = m_w;
	Rectangle* r = new Rectangle;
	Window window = XtWindow(w);
	if (window != 0)
	{
		XWindowAttributes a;
		Status s = XGetWindowAttributes(XtDisplay(w), window, &a);
		r->x = a.x;
		r->y = a.y;
		r->width = a.width;
		r->height = a.height;
	}
	else
	{
		XtWidgetGeometry a;
		XtGeometryResult result = XtMakeGeometryRequest(w, &a, NULL);
		r->x = a.x;
		r->y = a.y;
		r->width = a.width;
		r->height = a.height;
	}
	return r;
}

DimensionOb* Component::getSize()
{
	Widget w = m_w;
	DimensionOb* dim = new DimensionOb;
	Window window = XtWindow(w);
	if (window != 0)
	{
		XWindowAttributes a;
		Status s = XGetWindowAttributes(XtDisplay(w), window, &a);
		dim->width = a.width;
		dim->height = a.height;
	}
	else
	{
	}
	return dim;
}

int Component::getWidth()
{
	Widget w = m_w;
	Window window = XtWindow(w);
	int width;
	if (window != 0)
	{
		XWindowAttributes a;
		Status s = XGetWindowAttributes(XtDisplay(w), window, &a);
		width = a.width;
	}
	else
	{
		Arg arg;
		XtSetArg(arg, XtNwidth, &width);
		XtGetValues(w, &arg, 1);
	}
	return width;
}

int Component::getHeight()
{
	Widget w = m_w;
	Window window = XtWindow(w);
	int height;
	if (window != 0)
	{
		XWindowAttributes a;
		Status s = XGetWindowAttributes(XtDisplay(w), window, &a);
		height = a.height;
	}
	else
	{
		Arg arg;
		XtSetArg(arg, XtNheight, &height);
		XtGetValues(w, &arg, 1);
	}
	return height;
}

Color Component::getBackground()
{
	Color c;
	Arg arg;
	XtSetArg(arg, XtNbackground, &c.m_rgb);
	XtGetValues(m_w, &arg, 1);
	return c;
}

void Component::setBackground(Color c)
{
	XSetWindowBackground(XtDisplay(m_w), XtWindow(m_w), c.m_rgb);
}

Color Component::getForeground()
{
	Color c;
	Arg arg;
	XtSetArg(arg, XtNforeground, &c.m_rgb);
	XtGetValues(m_w, &arg, 1);
	return c.m_rgb;
}

void Component::setForeground(Color c)
{
//	XSetWindowForeground(XtDisplay(m_w), XtWindow(m_w), rgb);
}

void Component::setFont(Font font)
{
	Arg arg;
	XtSetArg(arg, XtNfont, &font);
	XtSetValues(m_w, &arg, 1);
}

// Create our standard XOR GC for drawing:
GC Component::getGC(int type)
{
	XGCValues values;
	Arg args[5];  // Leave room for future code modifications
	int n = 0;
	XtSetArg(args[n], XtNforeground, &values.foreground); n++;
	XtSetArg(args[n], XtNbackground, &values.background); n++;
//	XtSetArg(args[n], XtNfont, &values.font); n++;
	XtGetValues(m_w, args,n);

	values.line_style = LineSolid;
	// GXcopy, GXxor
	values.function   = type;
	GC gc = XtGetGC(m_w, GCForeground | GCBackground | //GCFont |
					GCFunction | GCLineStyle, &values);
	return gc;
}

Graphics* Component::getGraphics(Widget w)
{
	if (m_g == NULL)
	{
		m_g = new Graphics();
		m_g->m_w = w;
		m_g->m_gc = getGC(GXcopy); // GXxor
	}
	return m_g;
}

Graphics* Component::getGraphics()
{
	return getGraphics(m_w);
}

void Component::repaint(int time, int x, int y, int width, int height)
{
	// Erase area and generate a redraw/expose event:
	XClearArea(XtDisplay(m_w), XtWindow(m_w), x, y, width, height, 1);
}

void Component::repaint(int time)
{
	if (time == 0)
	{
		Graphics* g = getGraphics();
		update(g);
	}
	else
	{
		DimensionOb* dim = getSize();
		repaint(time, 0, 0, dim->width, dim->height);
		delete dim;
	}
}

void Component::update(Graphics* g)
{
	XClearWindow(XtDisplay(m_w), XtWindow(m_w));
	paint(g);
}

/////////////////////////////////////////////////////////////////////
// Container
Container::Container()
{
	m_layout = NULL;
}

Container::~Container()
{
	if (m_layout != NULL)
		XtDestroyWidget(m_layout);
}

/////////////////////////////////////////////////////////////////////
// Frame
Frame::Frame()
{
	m_menu = NULL;
}

Frame::~Frame()
{
	if (m_workWindow != NULL)
		XtDestroyWidget(m_workWindow);
	if (m_menu != NULL)
		delete m_menu;
}

void Frame::attach(Widget shell)
{
	m_w = shell;
	createLayout();
	if (m_workWindow != NULL)
		registerEventHandlers(m_workWindow);
	else
		registerEventHandlers(m_layout);
}

void Frame::createLayout()
{
#ifdef MOTIF
	m_layout = XmCreateMainWindow(m_w, "mw", NULL, 0);
	m_workWindow = XtCreateManagedWidget("", LABEL, m_layout, NULL, 0);
	XtVaSetValues(m_layout, XmNworkWindow, m_workWindow, NULL);
	XtManageChild(m_layout);
#else
	Arg args[2];
	int n = 0;
	XtSetArg(args[n], XtNborderWidth, 0); n++;
	m_mainWindow = XtCreateManagedWidget("form", FORM, m_shell, args, n);
#endif
}

Graphics* Frame::getGraphics()
{
	Widget w;
	if (m_workWindow != NULL)
		w = m_workWindow;
	else
		w = m_layout;
	return Component::getGraphics(w);
}

// Menu
void Frame::setMenuBar(char* menu_title, char** items, int num_items)
{
	m_menu_title = menu_title;
	m_menu_item_count = num_items;
	m_menu_items = items;

	// Menu
	MenuBar* menu = new MenuBar;
	menu->create(this);
	menu->addMenu(m_menu_title, m_menu_items, m_menu_item_count);
	m_menu = menu;
}

/////////////////////////////////////////////////////////////////////
// XApp
static char* s_display_server = "127.0.0.1:0.0";

XApp::XApp()
{
	m_appContext = NULL;
	m_display = NULL;
	m_shell = NULL;
	m_mainFrame = NULL;
	m_popup_active = 0;
}

XApp::~XApp()
{
	if (m_mainFrame != NULL)
		delete m_mainFrame;
	if (m_appContext != NULL)
		XtDestroyApplicationContext(m_appContext);
}

void XApp::setDisplayArg(int& argc, char**& argv)
{
	// Set -display option if not set
	static char* params[4];
	if (argc == 1)
	{
		printf("argc == 1\n");
		params[0] = argv[0];
		params[1] = "-display";
		params[2] = s_display_server;
		params[3] = NULL;
		argc = 3;
		argv = params;
	}
	printf("Option count: %d\n", argc);
	printf("%s\n", argv[1]);
	printf("%s\n", argv[2]);
}

void XApp::init(int argc, char **argv)
{
	setDisplayArg(argc, argv);
	// Initialize

	XtToolkitInitialize();
	m_appContext = XtCreateApplicationContext();
	char* app_name = NULL;
	char* app_class = "main";
	m_display = XtOpenDisplay(m_appContext, NULL, app_name, app_class,
								NULL, 0, &argc, argv);
	m_shell = XtAppCreateShell(app_name, app_class, APP_SHELL, m_display, NULL, 0);
}

void XApp::setMainFrame(Frame* frame)
{
	m_mainFrame = frame;
	m_mainFrame->attach(m_shell);
}

// Run
void XApp::run()
{
	XtRealizeWidget(m_shell);
	Widget w = m_mainFrame->m_w;
#ifndef MOTIF
	XGrabButton(XtDisplay(w), (unsigned int)AnyButton, AnyModifier,
				XtWindow(w), TRUE,
				(unsigned int)(ButtonPressMask | ButtonMotionMask | ButtonReleaseMask) ,
				GrabModeAsync, GrabModeAsync,
				XtWindow(w),
				XCreateFontCursor(XtDisplay(w), XC_crosshair));
#endif
	// Event Loop
	eventLoop();
}

void XApp::eventLoop()
{
	while (!XtAppGetExitFlag(m_appContext))
	{
		XEvent e;
		XtAppNextEvent(m_appContext, &e);
		XtDispatchEvent(&e);
	}
}

void XApp::exitEventLoop()
{
	XtAppSetExitFlag(m_appContext);
}
